@@ -1,5 +1,5 @@
-import sys
 import argparse
+import sys
 
 from parfive import Downloader, SessionConfig, __version__
 
@@ -39,15 +39,9 @@ def run_parfive(args):
 
 
 def parse_args(args):
-    parser = argparse.ArgumentParser(
-        description="Parfive: A parallel file downloader written in Python."
-    )
-    parser.add_argument(
-        "urls", metavar="URLS", type=str, nargs="+", help="URLs of files to be downloaded."
-    )
-    parser.add_argument(
-        "--max-conn", type=int, default=5, help="Maximum number of parallel file downloads."
-    )
+    parser = argparse.ArgumentParser(description="Parfive: A parallel file downloader written in Python.")
+    parser.add_argument("urls", metavar="URLS", type=str, nargs="+", help="URLs of files to be downloaded.")
+    parser.add_argument("--max-conn", type=int, default=5, help="Maximum number of parallel file downloads.")
     parser.add_argument(
         "--max-splits",
         type=int,
@@ -97,5 +91,4 @@ def parse_args(args):
     )
     parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
 
-    args = parser.parse_args(args)
-    return args
+    return parser.parse_args(args)
