@@ -2,19 +2,17 @@ import os
 import platform
 import threading
 from pathlib import Path
-from tempfile import gettempdir
 from unittest import mock
 from unittest.mock import MagicMock, patch
 
 import aiohttp
 import pytest
 from aiohttp import ClientConnectorError, ClientTimeout
-from pytest_socket import SocketConnectBlockedError
 
 import parfive
 from parfive.config import SessionConfig
 from parfive.downloader import Downloader, FailedDownload, Results, Token
-from parfive.utils import sha256sum
+from parfive.utils import ChecksumMismatch, check_file_hash
 
 skip_windows = pytest.mark.skipif(platform.system() == "Windows", reason="Windows.")
 
@@ -22,7 +20,10 @@ skip_windows = pytest.mark.skipif(platform.system() == "Windows", reason="Window
 def validate_test_file(f):
     assert len(f) == 1
     assert Path(f[0]).name == "testfile.fits"
-    assert sha256sum(f[0]) == "a1c58cd340e3bd33f94524076f1fa5cf9a7f13c59d5272a9d4bc0b5bc436d9b3"
+    with Path(f[0]).open(mode="rb") as fobj:
+        assert check_file_hash(
+            fobj, "sha-256=a1c58cd340e3bd33f94524076f1fa5cf9a7f13c59d5272a9d4bc0b5bc436d9b3"
+        )
 
 
 def test_setup():
@@ -47,6 +48,8 @@ def test_download(httpserver, tmpdir):
     assert dl.queued_downloads == 1
 
     f = dl.download()
+    assert len(f.urls) == 1
+    assert f.urls[0] == httpserver.url
     validate_test_file(f)
 
 
@@ -245,7 +248,7 @@ def test_retrieve_some_content(testserver, tmpdir):
 
     nn = 5
     for i in range(nn):
-        dl.enqueue_file(testserver.url, path=tmpdir)
+        dl.enqueue_file(testserver.url + f"/testfile_{i}.txt", path=tmpdir)
 
     f = dl.download()
 
@@ -302,7 +305,10 @@ def test_failed_download():
 def test_results():
     res = Results()
 
-    res.append("hello")
+    res.append(path="hello", url="aurl")
+
+    assert res[0] == "hello"
+    assert res.urls[0] == "aurl"
 
     res.add_error("wibble", "notaurl", "out of cheese")
 
@@ -339,7 +345,7 @@ def test_retry(tmpdir, testserver):
 
     nn = 5
     for i in range(nn):
-        dl.enqueue_file(testserver.url, path=tmpdir)
+        dl.enqueue_file(testserver.url + f"/testfile_{i}.txt", path=tmpdir)
 
     f = dl.download()
 
@@ -350,6 +356,10 @@ def test_retry(tmpdir, testserver):
 
     assert len(f2) == nn
     assert len(f2.errors) == 0
+    assert [f"testfile_{k}.txt" for k in ["0", "1", "3", "4", "2"]] == [f.split("/")[-1] for f in f2.urls]
+    assert "testfile_0.txt" == Path(f2[0]).name
+    assert "testfile_4.txt" == Path(f2[-2]).name
+    assert "testfile_2.txt" == Path(f2[-1]).name
 
 
 def test_empty_retry():
@@ -359,23 +369,20 @@ def test_empty_retry():
     dl.retry(f)
 
 
-def test_done_callback_error(tmpdir, testserver):
-    tmpdir = str(tmpdir)
-
+def test_done_callback_error(tmp_path, testserver):
     def done_callback(filepath, url, error):
         if error is not None:
-            (Path(gettempdir()) / "callback.error").touch()
+            (tmp_path / "callback.error").touch()
 
     dl = Downloader(config=SessionConfig(done_callbacks=[done_callback]))
 
-    nn = 5
-    for i in range(nn):
-        dl.enqueue_file(testserver.url, path=tmpdir)
+    dl.enqueue_file(testserver.url + "/testfile_2.txt", path=tmp_path)
 
     f = dl.download()
 
-    assert (Path(gettempdir()) / "callback.error").exists()
-    (Path(gettempdir()) / "callback.error").unlink()
+    assert len(f.errors) == 1
+
+    assert (tmp_path / "callback.error").exists()
 
 
 @skip_windows
@@ -394,6 +401,8 @@ def test_ftp(tmpdir):
     assert len(f.errors) == 3
 
 
+# I don't know of an alternative server which makes any sense to test this with
+@pytest.mark.skip("Remote server offline")
 @skip_windows
 @pytest.mark.allow_hosts(True)
 def test_ftp_pasv_command(tmpdir):
@@ -470,7 +479,7 @@ def test_custom_user_agent(httpserver, tmpdir):
 
 @patch.dict(os.environ, {"HTTP_PROXY": "http_proxy_url", "HTTPS_PROXY": "https_proxy_url"})
 @pytest.mark.parametrize(
-    "url,proxy",
+    ("url", "proxy"),
     [
         ("http://test.example.com", "http_proxy_url"),
         ("https://test.example.com", "https_proxy_url"),
@@ -487,11 +496,12 @@ def test_proxy_passed_as_kwargs_to_get(tmpdir, url, proxy):
 
     assert patched.called, "`ClientSession._request` not called"
     assert list(patched.call_args) == [
-        ("GET", url),
+        ("HEAD", url),
         {
-            "allow_redirects": True,
+            "allow_redirects": False,
             "timeout": ClientTimeout(total=0, connect=None, sock_read=90, sock_connect=None),
             "proxy": proxy,
+            "headers": {},
         },
     ]
 
@@ -517,11 +527,11 @@ def test_http_callback_success(httpserver, tmpdir):
     assert cb_status is None
 
 
-def test_http_callback_fail(httpserver, tmpdir):
+def test_http_callback_fail(tmpdir):
     # Test callback on failed download
     cb = MagicMock()
     dl = Downloader(config=SessionConfig(done_callbacks=[cb]))
-    url = "http://test.com/myfile.txt"
+    url = "http://127.0.0.1/myfile.txt"
     dl.enqueue_file(url, path=tmpdir, max_splits=None)
 
     assert dl.queued_downloads == 1
@@ -532,7 +542,8 @@ def test_http_callback_fail(httpserver, tmpdir):
     cb_path, cb_url, cb_status = cb.call_args[0]
     assert cb_path is None
     assert url == cb_url
-    assert isinstance(cb_status, (SocketConnectBlockedError, ClientConnectorError))
+    # Returns 404 on windows on GHA, which triggers FailedDownload
+    assert isinstance(cb_status, (ClientConnectorError, FailedDownload))
 
 
 @pytest.mark.allow_hosts(True)
@@ -603,9 +614,258 @@ def test_download_out_of_main_thread(httpserver, tmpdir, recwarn):
     # We use recwarn here as for some reason pytest.warns did not reliably pickup this warning.
     assert len(recwarn) > 0
     assert any(
-        [
-            "This download has been started in a thread which is not the main thread. You will not be able to interrupt the download."
-            == w.message.args[0]
-            for w in recwarn
-        ]
+        "This download has been started in a thread which is not the main thread. You will not be able to interrupt the download."
+        == w.message.args[0]
+        for w in recwarn
     )
+
+
+def test_checksum_want_headers(httpserver, tmpdir):
+    tmpdir = str(tmpdir)
+    httpserver.serve_content(
+        "SIMPLE  = T",
+        headers={
+            "Content-Disposition": "attachment; filename=testfile.fits",
+            "Repr-Digest": "sha-256=a1c58cd340e3bd33f94524076f1fa5cf9a7f13c59d5272a9d4bc0b5bc436d9b3",
+        },
+    )
+    dl = Downloader()
+
+    dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None, checksum=True)
+    dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None, checksum=False)
+
+    assert dl.queued_downloads == 2
+
+    f = dl.download()
+    assert len(f.urls) == 2
+    assert f.urls[0] == httpserver.url
+    validate_test_file(f[0:1])
+
+    first_headers = httpserver.requests[0].headers
+    assert "Want-Repr-Digest" in first_headers
+    assert "Want-Content-Digest" in first_headers
+
+    # Two requests a file
+    second_headers = httpserver.requests[2].headers
+    assert "Want-Repr-Digest" not in second_headers
+    assert "Want-Content-Digest" not in second_headers
+
+
+def test_checksum_invalid(httpserver, tmpdir):
+    tmpdir = str(tmpdir)
+    httpserver.serve_content(
+        "SIMPLE  = T",
+        headers={
+            "Content-Disposition": "attachment; filename=testfile.fits",
+            "Content-Digest": "sha-256=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
+        },
+    )
+    dl = Downloader()
+
+    dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None, checksum=True)
+
+    assert dl.queued_downloads == 1
+
+    f = dl.download()
+
+    assert len(f.errors) == 1
+    exception = f.errors[0].exception
+    assert isinstance(exception, FailedDownload)
+    assert "checksum doesn't match" in str(exception)
+
+
+def test_early_fail_download_checksum_mismatch(httpserver, tmpdir):
+    tmpdir = str(tmpdir)
+    httpserver.serve_content(
+        "SIMPLE  = T",
+        headers={
+            "Content-Disposition": "attachment; filename=testfile.fits",
+            "Repr-Digest": "sha-256=INVALID",
+        },
+    )
+    dl = Downloader()
+
+    dl.enqueue_file(
+        httpserver.url,
+        path=Path(tmpdir),
+        max_splits=None,
+        checksum="sha-256=a1c58cd340e3bd33f94524076f1fa5cf9a7f13c59d5272a9d4bc0b5bc436d9b3",
+    )
+
+    assert dl.queued_downloads == 1
+
+    f = dl.download()
+
+    assert len(f.errors) == 1
+    exc = f.errors[0].exception
+    assert isinstance(exc, FailedDownload)
+    assert isinstance(exc.exception, ChecksumMismatch)
+    assert "Server provided checksum and user provided checksum do not match, download skipped" in str(
+        exc.exception
+    )
+
+
+def test_server_user_algorithm_mismatch(httpserver, tmp_path, caplog):
+    caplog.set_level("INFO")
+    httpserver.serve_content(
+        "SIMPLE  = T",
+        headers={
+            "Content-Disposition": "attachment; filename=testfile.fits",
+            "Repr-Digest": "sha-512=INVALID",
+        },
+    )
+    dl = Downloader()
+
+    dl.enqueue_file(
+        httpserver.url,
+        path=tmp_path,
+        max_splits=None,
+        checksum="sha-256=a1c58cd340e3bd33f94524076f1fa5cf9a7f13c59d5272a9d4bc0b5bc436d9b3",
+    )
+
+    assert dl.queued_downloads == 1
+
+    f = dl.download()
+
+    assert len(f.errors) == 0
+
+    assert any(
+        "Not comparing user provided checksum to server provided checksum as algorithms do not match (got sha512 from the server)."
+        in m
+        for m in caplog.messages
+    )
+
+    first_headers = httpserver.requests[0].headers
+    assert "Want-Repr-Digest" in first_headers
+    assert "Want-Content-Digest" in first_headers
+    assert "sha-256=10" in first_headers["Want-Repr-Digest"]
+    assert "sha-256=10" in first_headers["Want-Content-Digest"]
+
+
+def test_explicit_checksum(namedserver, tmpdir):
+    tmpdir = str(tmpdir)
+    dl = Downloader()
+
+    dl.enqueue_file(
+        namedserver.url + "/testfile.txt",
+        path=Path(tmpdir),
+        max_splits=None,
+        checksum="sha-256=0ba904eae8773b70c75333db4de2f3ac45a8ad4ddba1b242f0b3cfc199391dd8",
+    )
+
+    dl.enqueue_file(
+        namedserver.url + "/testfile.txt",
+        path=Path(tmpdir),
+        max_splits=None,
+        checksum="sha-256=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
+    )
+
+    assert dl.queued_downloads == 2
+
+    f = dl.download()
+
+    assert len(f) == 1
+    assert f.urls[0] == namedserver.url + "/testfile.txt"
+
+    assert len(f.errors) == 1
+    exception = f.errors[0].exception
+    assert isinstance(exception, FailedDownload)
+    assert "checksum doesn't match" in str(exception)
+
+
+def test_file_exists_checksum(httpserver, tmpdir, caplog):
+    caplog.set_level("DEBUG")
+    tmpdir = str(tmpdir)
+    httpserver.serve_content(
+        "SIMPLE  = T",
+        headers={
+            "Content-Disposition": "attachment; filename=testfile.fits",
+            "Repr-Digest": "sha-256=a1c58cd340e3bd33f94524076f1fa5cf9a7f13c59d5272a9d4bc0b5bc436d9b3",
+        },
+    )
+    dl = Downloader()
+
+    dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None, checksum=True)
+
+    assert dl.queued_downloads == 1
+
+    f = dl.download()
+
+    assert len(f) == 1
+
+    dl = Downloader()
+
+    dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None, checksum=True)
+
+    assert dl.queued_downloads == 1
+
+    f = dl.download()
+
+    assert len(f) == 1
+
+    # Assert that only three requests have been made, thereby checking
+    # that the file wasn't re-downloaded the second time
+    assert len(httpserver.requests) == 3
+
+    assert (
+        "testfile.fits already exists, checksum matches and overwrite is False; skipping download."
+        in caplog.messages[-1]
+    )
+
+
+def test_no_server_checksum(httpserver, tmpdir, caplog):
+    caplog.set_level("DEBUG")
+    tmpdir = str(tmpdir)
+    httpserver.serve_content(
+        "SIMPLE  = T",
+        headers={
+            "Content-Disposition": "attachment; filename=testfile.fits",
+        },
+    )
+    dl = Downloader()
+
+    dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None, checksum=True)
+
+    assert dl.queued_downloads == 1
+
+    f = dl.download()
+
+    assert len(f) == 1
+
+    assert any("Expected server to provide checksum for url" in msg for msg in caplog.messages)
+
+
+def test_invalid_checksum_enqueue():
+    dl = Downloader()
+
+    with pytest.raises(ValueError, match="checksum 'wibble' should be of the format <algorithm>=<checksum>"):
+        dl.enqueue_file("", checksum="wibble")
+
+    with pytest.raises(ValueError, match="checksum type 'nope' is not supported"):
+        dl.enqueue_file("", checksum="nope=wibble")
+
+    with pytest.raises(ValueError, match="checksum type 'nope' is not supported"):
+        check_file_hash("", "nope=wibble")
+
+
+@pytest.mark.parametrize("checksum", ["nope=wibble", "wibble"])
+def test_invalid_server_checksum(httpserver, tmpdir, caplog, checksum):
+    caplog.set_level("ERROR")
+    tmpdir = str(tmpdir)
+    httpserver.serve_content(
+        "SIMPLE  = T",
+        headers={
+            "Content-Disposition": "attachment; filename=testfile.fits",
+            "Content-Digest": checksum,
+        },
+    )
+    dl = Downloader()
+
+    dl.enqueue_file(httpserver.url, path=Path(tmpdir), max_splits=None, checksum=True)
+
+    assert dl.queued_downloads == 1
+
+    f = dl.download()
+
+    assert len(f.errors) == 0
+    assert "Got invalid checksum:" in caplog.messages[0]
