@@ -1,12 +1,16 @@
-import os
 import asyncio
 import hashlib
+import io
+import os
 import pathlib
+import typing
 import warnings
-from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union, TypeVar, Generator
-from pathlib import Path
-from itertools import count
+from collections.abc import AsyncIterator, Generator
 from concurrent.futures import ThreadPoolExecutor
+from contextlib import asynccontextmanager
+from itertools import count
+from pathlib import Path
+from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar, Union
 
 import aiohttp
 
@@ -17,10 +21,9 @@ if TYPE_CHECKING:
 
 
 __all__ = [
-    "cancel_task",
-    "run_in_thread",
-    "Token",
     "FailedDownload",
+    "Token",
+    "cancel_task",
     "default_name",
     "remove_file",
 ]
@@ -40,7 +43,7 @@ def _parseparam(s: str) -> Generator[str, None, None]:
         s = s[end:]
 
 
-def parse_header(line: str) -> Tuple[str, Dict[str, str]]:
+def parse_header(line: str) -> tuple[str, dict[str, str]]:
     """Parse a Content-type like header.
     Return the main content-type and a dictionary of options.
     """
@@ -97,7 +100,7 @@ async def get_ftp_size(client: "aioftp.Client", filepath: os.PathLike) -> int:
     try:
         size = await client.stat(filepath)
         size = size.get("size", None)
-    except Exception:
+    except Exception:  # noqa BLE001
         parfive.log.info("Failed to get size of FTP file", exc_info=True)
         size = None
 
@@ -130,7 +133,7 @@ def replacement_filename(path: os.PathLike) -> Path:  # type: ignore[return]
             return new_path
 
 
-def get_filepath(filepath: os.PathLike, overwrite: bool) -> Tuple[Union[Path, str], bool]:
+def get_filepath(filepath: os.PathLike, overwrite: Union[bool, Literal["unique"]]) -> tuple[Path, bool]:
     """
     Get the filepath to download to and ensure dir exists.
 
@@ -141,7 +144,7 @@ def get_filepath(filepath: os.PathLike, overwrite: bool) -> Tuple[Union[Path, st
     filepath = pathlib.Path(filepath)
     if filepath.exists():
         if not overwrite:
-            return str(filepath), True
+            return filepath, True
         if overwrite == "unique":
             filepath = replacement_filename(filepath)
     if not filepath.parent.exists():
@@ -150,26 +153,13 @@ def get_filepath(filepath: os.PathLike, overwrite: bool) -> Tuple[Union[Path, st
     return filepath, False
 
 
-def sha256sum(filename: str) -> str:
-    """
-    https://stackoverflow.com/a/44873382
-    """
-    h = hashlib.sha256()
-    b = bytearray(128 * 1024)
-    mv = memoryview(b)
-    with open(filename, "rb", buffering=0) as f:
-        for n in iter(lambda: f.readinto(mv), 0):
-            h.update(mv[:n])
-    return h.hexdigest()
-
-
 class MultiPartDownloadError(Exception):
     def __init__(self, response: aiohttp.ClientResponse) -> None:
         self.response = response
 
 
 class FailedDownload(Exception):
-    def __init__(self, filepath_partial: Path, url: str, exception: BaseException) -> None:
+    def __init__(self, filepath_partial: Union[Path, Callable], url: str, exception: BaseException) -> None:
         self.filepath_partial = filepath_partial
         self.url = url
         self.exception = exception
@@ -181,7 +171,11 @@ class FailedDownload(Exception):
         return out
 
     def __str__(self) -> str:
-        return f"Download Failed: {self.url} with error {str(self.exception)}"
+        return f"Download Failed: {self.url} with error {self.exception!s}"
+
+
+class ChecksumMismatch(Exception):
+    """Used when a checksum doesn't match."""
 
 
 class Token:
@@ -198,7 +192,7 @@ class Token:
 _T = TypeVar("_T")
 
 
-class _QueueList(List[_T]):
+class _QueueList(list[_T]):
     """
     A list, with an extra method that empties the list and puts it into a
     `asyncio.Queue`.
@@ -233,7 +227,7 @@ def remove_file(filepath: os.PathLike) -> None:
     filepath = Path(filepath)
     try:
         filepath.unlink(missing_ok=True)
-    except Exception as remove_exception:
+    except Exception as remove_exception:  # noqa BLE001
         warnings.warn(
             f"Failed to delete possibly incomplete file {filepath} {remove_exception}",
             ParfiveUserWarning,
@@ -252,3 +246,97 @@ async def cancel_task(task: asyncio.Task) -> bool:
     except asyncio.CancelledError:
         return True
     return task.cancelled()
+
+
+try:
+    # Python 3.11 added file_digest
+    from hashlib import file_digest
+except ImportError:
+    import hashlib
+
+    # Copied from the stdlib
+    @typing.no_type_check
+    def file_digest(fileobj, digest, /, *, _bufsize=2**18):
+        """
+        Hash the contents of a file-like object. Returns a digest object.
+
+        *fileobj* must be a file-like object opened for reading in binary mode.
+        It accepts file objects from open(), io.BytesIO(), and SocketIO objects.
+        The function may bypass Python's I/O and use the file descriptor *fileno*
+        directly.
+
+        *digest* must either be a hash algorithm name as a *str*, a hash
+        constructor, or a callable that returns a hash object.
+        """
+        # On Linux we could use AF_ALG sockets and sendfile() to archive zero-copy
+        # hashing with hardware acceleration.
+        if isinstance(digest, str):
+            digestobj = hashlib.new(digest)
+        else:
+            digestobj = digest()
+
+        if hasattr(fileobj, "getbuffer"):
+            # io.BytesIO object, use zero-copy buffer
+            digestobj.update(fileobj.getbuffer())
+            return digestobj
+
+        # Only binary files implement readinto().
+        if not (hasattr(fileobj, "readinto") and hasattr(fileobj, "readable") and fileobj.readable()):
+            raise ValueError(f"'{fileobj!r}' is not a file-like object in binary reading mode.")
+
+        # binary file, socket.SocketIO object
+        # Note: socket I/O uses different syscalls than file I/O.
+        buf = bytearray(_bufsize)  # Reusable buffer to reduce allocations.
+        view = memoryview(buf)
+        while True:
+            size = fileobj.readinto(buf)
+            if size is None:
+                raise BlockingIOError("I/O operation would block.")
+            if size == 0:
+                break  # EOF
+            digestobj.update(view[:size])
+
+        return digestobj
+
+
+def validate_checksum_format(checksum: str) -> tuple[str, str]:
+    if "=" not in checksum:
+        raise ValueError(f"checksum '{checksum}' should be of the format <algorithm>=<checksum>")
+    chk_alg, checksum = checksum.split("=")
+    # Normalise the algorithm name to not have "-" as that might have wider support
+    chk_alg = chk_alg.replace("-", "")
+    if chk_alg not in hashlib.algorithms_available:
+        raise ValueError(f"checksum type '{chk_alg}' is not supported.")
+    return chk_alg, checksum
+
+
+def check_file_hash(fileobj: io.BufferedReader, checksum: str, accept_invalid_checksum: bool = False) -> bool:
+    """
+    Verify the contents of fileobj match the checksum provided by ``checksum``.
+    """
+    try:
+        chk_alg, checksum = validate_checksum_format(checksum)
+    except ValueError as e:
+        if not accept_invalid_checksum:
+            raise
+        parfive.log.error("Got invalid checksum: %s", e)
+        # Allow invalid checksums to match
+        return True
+
+    computed_file_hash = file_digest(fileobj, chk_alg).hexdigest()
+    return computed_file_hash == checksum
+
+
+@asynccontextmanager
+async def session_head_or_get(session: aiohttp.ClientSession, url: str, **kwargs: dict) -> AsyncIterator:
+    """
+    Try and make a HEAD request to the resource and fallback to a get
+    request if that fails.
+    """
+    async with session.head(url, **kwargs) as resp:
+        if resp.status == 200:
+            yield resp
+            return
+
+    async with session.get(url, **kwargs) as resp:
+        yield resp
